package org.GNOME.Speech;
import org.GNOME.Bonobo.*;
import org.GNOME.Speech.*;
import org.GNOME.Accessibility.*;
import org.omg.CORBA.*;
import java.util.Vector;
import java.io.*;
import com.sun.speech.freetts.Voice;
import com.sun.speech.freetts.en.us.CMULexicon;
import com.sun.speech.freetts.audio.JavaStreamingAudioPlayer;
import com.sun.speech.freetts.audio.JavaClipAudioPlayer;

public class FreeTTSSynthesisDriver extends UnknownImpl implements SynthesisDriverOperations {

	private boolean initialized = false;
	private Voice v = null;
	private CMULexicon l = null;
	private int index = 0;
	private SpeechThread thread = null;
	private float defaultPitch;
	private float defaultRate;
	private float defaultVolume;
	
	
	public void bonoboActivate (String iid, String [] args) {
		ORB orb = AccessUtil.getORB ();
		String ior = orb.object_to_string (tie._this_object());
		_bonoboActivate (ior, iid, args);
	}

	public native void _bonoboActivate (String ior, String iid,
					    String [] arggs);

	static {
		System.loadLibrary ("freettsdriver");
	}
    
	public FreeTTSSynthesisDriver (String [] args) {
		poa = JavaBridge.getRootPOA ();
		tie = new SynthesisDriverPOATie (this, poa);
		try {
			poa.activate_object(tie);
			bonoboActivate ("OAFIID:GNOME_Speech_SynthesisDriver_FreeTTS:proto0.2", args);
		} catch (Exception e) {
			System.err.println ("error: " + e);
		}
	}

	/* Descriptive attributes */

	public String driverName () {
		return "FreeTTS Gnome Speech Driver";
	}

	public String synthesizerName () {
		return "FreeTTS";
	}

	public String driverVersion () {
		return "0.2";
	}
    
	public String synthesizerVersion () {
		return "1.1.2";
	}

	/* Initialization */

	public boolean driverInit () {
		if (initialized)
			return true;
		System.setProperty
			("com.sun.speech.freetts.audio.AudioPlayer.cancelDelay", "0");
		String databaseName = System.getProperty
			("diphoneDatabase", "cmu_kal/diphone_units16.bin");
		v = (Voice) new com.sun.speech.freetts.en.us.CMUDiphoneVoice ();
		v.getFeatures().setString
			(Voice.DATABASE_NAME, databaseName);	
		l = new CMULexicon ();
		v.setLexicon (l);
		v.load ();
		v.setAudioPlayer (new JavaClipAudioPlayer());

		/* Set default speech parameters */

		defaultPitch = v.getPitch();
		defaultRate = v.getRate();
		defaultVolume = v.getVolume();
		
		thread = new SpeechThread(v);
		thread.start ();
		initialized = true;
		return true;
	}

	public boolean isInitialized () {
		return initialized;
	}	

	/* Voice related functions */

	public org.GNOME.Speech.VoiceInfo[] getVoices (org.GNOME.Speech.VoiceInfo voice_spec) {
		return getAllVoices ();
	}

	public org.GNOME.Speech.VoiceInfo[] getAllVoices () {
		VoiceInfo[] voices = new VoiceInfo[1];		
		voices[0] = new VoiceInfo ("default",
					   (short)0, voice_gender.gender_male);
		return voices;
	}

	/* Speaker creation */

	public org.GNOME.Speech.Speaker createSpeaker (org.GNOME.Speech.VoiceInfo voice_spec) {
		Speaker speaker;

		speaker = SpeakerHelper.narrow ((new FreeTTSSpeaker (this)).tie());
		return speaker;
	}

	public static void main (String [] args) {
		try {
			FreeTTSSynthesisDriver driver = new FreeTTSSynthesisDriver (args);	    
			AccessUtil.getORB().run();
		} catch (Exception e) {
			System.err.println ("error: " + e);
		}
	}
	
	public int say (FreeTTSSpeaker s,
			String text) {
		index++;
		thread.addEntry (s, index, text);
		return index;
	}
	
	public boolean stop () {
		thread.removeAll ();
		return true;
	}

	public boolean isSpeaking () {
		return true;
	}

	public Voice getVoice () {
		return v;
	}

	public float getDefaultPitch () {
		return defaultPitch;
	}

	public float getDefaultRate () {
		return defaultRate;
	}

	public float getDefaultVolume () {
		return defaultVolume;
	}

	public void unref () {
		if (thread != null)
			thread.quit ();
		super.unref ();
	}

	class QueueEntry {

		public int index;
		public String text;
		FreeTTSSpeaker s = null;

		public QueueEntry (FreeTTSSpeaker s, 
				   int index,
				   String text) {
			this.s = s;
			this.text = text;
			this.index = index;
		}
		
	}
	
	class SpeechThread extends Thread {
		private boolean done = false;
		Voice v;
		private Vector SpeechQueue = new Vector ();
	
		public SpeechThread (Voice v) {
			this.v = v;
		}
	
		public void run () {
			while (done == false) {
				QueueEntry e = null;
				synchronized (SpeechQueue) {
					if (SpeechQueue.size() == 0) {
						try {
							SpeechQueue.wait ();
						} catch (InterruptedException ie) {
						}
					}
				}
				synchronized (SpeechQueue) {
					e = (QueueEntry) SpeechQueue.remove(0);
				}
				if (e != null) {
					
					/* Set speaker parameters */

					if (e.s.needsParameterRefresh()) {
						v.setPitch (e.s.getPitch());
						v.setRate(e.s.getRate());
						v.setVolume(e.s.getVolume());
					}
					if (e.s.getCallback() != null)
						e.s.getCallback()._notify (speech_callback_type.speech_callback_speech_started,
							      e.index, 0);
					v.speak (e.text);
					if (e.s.getCallback() != null)
						e.s.getCallback()._notify (speech_callback_type.speech_callback_speech_ended,
							    e.index, 0);
				}
			}
		}
	
		public void addEntry (FreeTTSSpeaker s,
				      int index,
				      String text) {
			QueueEntry e = new QueueEntry (s, index, text);
			synchronized (SpeechQueue) {
				SpeechQueue.add(e);
				SpeechQueue.notifyAll();
			}
		}


		public synchronized void quit () {
			removeAll ();
			done = true;
		}

		public synchronized void removeAll () {
			v.getAudioPlayer().cancel();
			synchronized (SpeechQueue) {
				SpeechQueue.removeAllElements();
			}
		}
	}
}
